import copy
import enum
import re
from typing import Iterable, Iterator, Optional
from xml.etree.ElementTree import Element, SubElement, tostring
from xml.sax.saxutils import unescape

from latex2mathml import commands
from latex2mathml.symbols_parser import convert_symbol
from latex2mathml.walker import MULTIPRIMES, Node, walk

COLUMN_ALIGNMENT_MAP = {"r": "right", "l": "left", "c": "center"}
OPERATORS = frozenset(
    (
        "+",
        "-",
        "*",
        "/",
        "(",
        ")",
        "=",
        ",",
        "?",
        "[",
        "]",
        "|",
        r"\|",
        "!",
        r"\{",
        r"\}",
        r">",
        r"<",
        r".",
        r"\ast",
        r"\bigotimes",
        r"\cdot",
        r"\centerdot",
        r"\div",
        r"\dots",
        r"\dotsc",
        r"\dotso",
        r"\gt",
        r"\ldotp",
        r"\lt",
        r"\lvert",
        r"\lVert",
        r"\lvertneqq",
        r"\ngeqq",
        r"\omicron",
        r"\rvert",
        r"\rVert",
        r"\S",
        r"\smallfrown",
        r"\smallint",
        r"\smallsmile",
        r"\surd",
        r"\times",
        r"\varsubsetneqq",
        r"\varsupsetneqq",
    )
)
MATH_MODE_PATTERN = re.compile(r"\\\$|\$|\\?[^\\$]+")
NUMBER_PATTERN = re.compile(r"\d+(\.\d+)?")
MOVABLE_LIMIT_TEXTS = {
    commands.ARGMAX: "arg&#x02009;max",
    commands.ARGMIN: "arg&#x02009;min",
    commands.DETERMINANT: "det",
    commands.GCD: "gcd",
    commands.INJLIM: "inj&#x02006;lim",
    commands.INTOP: "&#x0222B;",
    commands.LIMINF: "lim&#x02006;inf",
    commands.LIMSUP: "lim&#x02006;sup",
    commands.PLIM: "plim",
    commands.PR: "Pr",
    commands.PROJLIM: "proj&#x02006;lim",
    commands.VARINJLIM: "inj&#x02006;lim",
    commands.VARLIMINF: "lim&#x02006;inf",
    commands.VARLIMSUP: "lim&#x02006;sup",
    commands.VARPROJLIM: "proj&#x02006;lim",
}


class Mode(enum.Enum):
    TEXT = enum.auto()
    MATH = enum.auto()


class Converter:
    def __init__(self, xmlns: str = "http://www.w3.org/1998/Math/MathML", display: str = "inline") -> None:
        self.xmlns = xmlns
        self.display = display
        self.equation_counter = 0
        self.macros: dict[str, tuple[list[str], int]] = {}

    def convert(self, latex: str, parent: Optional[Element] = None) -> str:
        return self._convert(self.convert_to_element(latex, parent))

    def convert_to_element(self, latex: str, parent: Optional[Element] = None) -> Element:
        tag = "math"
        attrib = {"xmlns": self.xmlns, "display": self.display}
        math = Element(tag, attrib) if parent is None else SubElement(parent, tag, attrib)
        row = SubElement(math, "mrow")
        self._convert_group(iter(walk(latex, self.display, macros=self.macros)), row)
        return math

    def reset(self) -> None:
        self.equation_counter = 0
        self.macros = {}

    @staticmethod
    def _convert(tree: Element) -> str:
        return unescape(tostring(tree, encoding="unicode"))

    def _convert_matrix(
        self, nodes: Iterator[Node], parent: Element, command: str, alignment: Optional[str] = None
    ) -> None:
        row = None
        cell = None

        col_index = 0
        col_alignment = None

        max_col_size = 0

        row_index = 0
        row_lines: list[str] = []

        hfil_indexes: list[bool] = []

        numbered = command == commands.ALIGN
        skip_number = False

        for node in nodes:
            if row is None:
                row = SubElement(parent, "mtr")

            if cell is None:
                col_alignment, col_index = self._get_column_alignment(alignment, col_alignment, col_index)
                cell = self._make_matrix_cell(row, col_alignment)

            if node.token == commands.BRACES:
                self._convert_group(iter([node]), cell)
            elif node.token == "&":
                self._set_cell_alignment(cell, hfil_indexes)
                hfil_indexes = []
                col_alignment, col_index = self._get_column_alignment(alignment, col_alignment, col_index)
                cell = self._make_matrix_cell(row, col_alignment)
                if command in (commands.SPLIT, commands.ALIGN, commands.ALIGNSTAR) and col_index % 2 == 0:
                    SubElement(cell, "mi")
            elif node.token in (commands.DOUBLEBACKSLASH, commands.CARRIAGERETURN):
                self._set_cell_alignment(cell, hfil_indexes)
                hfil_indexes = []
                if numbered and not skip_number:
                    self.equation_counter += 1
                    eqn_cell = SubElement(row, "mtd")
                    eqn_num = SubElement(eqn_cell, "mtext")
                    eqn_num.text = f"({self.equation_counter})"
                skip_number = False
                row_index += 1
                if col_index > max_col_size:
                    max_col_size = col_index
                col_index = 0
                col_alignment, col_index = self._get_column_alignment(alignment, col_alignment, col_index)
                row = SubElement(parent, "mtr")
                cell = self._make_matrix_cell(row, col_alignment)
            elif node.token in (commands.NONUMBER, commands.NOTAG):
                skip_number = True
            elif node.token == commands.HLINE:
                row_lines.append("solid")
            elif node.token == commands.HDASHLINE:
                row_lines.append("dashed")
            elif node.token == commands.HFIL:
                hfil_indexes.append(True)
            else:
                if row_index > len(row_lines):
                    row_lines.append("none")
                hfil_indexes.append(False)
                self._convert_group(iter([node]), cell)

        if col_index > max_col_size:
            max_col_size = col_index

        if any(r != "none" for r in row_lines):
            parent.set("rowlines", " ".join(row_lines))

        if row is not None and cell is not None and len(cell) == 0:
            parent.remove(row)
            row = None

        if numbered and row is not None and not skip_number:
            self.equation_counter += 1
            eqn_cell = SubElement(row, "mtd")
            eqn_num = SubElement(eqn_cell, "mtext")
            eqn_num.text = f"({self.equation_counter})"

        if max_col_size and command in (commands.ALIGN, commands.ALIGNSTAR):
            spacing = ("0em", "2em")
            multiplier = max_col_size // len(spacing)
            parent.set("columnspacing", " ".join(spacing * multiplier))

    def _convert_group(
        self, nodes: Iterable[Node], parent: Element, font: Optional[dict[str, Optional[str]]] = None
    ) -> None:
        _font = font
        for node in nodes:
            token = node.token
            if token in (*commands.MSTYLE_SIZES, *commands.STYLES):
                node = Node(token=token, children=tuple(n for n in nodes))
                self._convert_command(node, parent, _font)
            elif token == commands.UNICODE:
                arg = node.children[0] if node.children else None
                if arg and arg.children:
                    code = "".join(c.token for c in arg.children)
                elif arg:
                    code = arg.token
                else:
                    code = ""
                element = SubElement(parent, "mi")
                element.text = f"&#x{code.lstrip('x')};"
            elif token == commands.RULE:
                attrs = node.attributes or {}
                SubElement(parent, "mspace", mathbackground="black", width=attrs["width"], height=attrs["height"])
            elif token in commands.CONVERSION_MAP or token in (commands.MOD, commands.PMOD, commands.POD):
                self._convert_command(node, parent, _font)
            elif token in commands.LOCAL_FONTS and node.children is not None:
                self._convert_group(iter(node.children), parent, commands.LOCAL_FONTS[token])
            elif token.startswith(commands.MATH) and node.children is not None:
                self._convert_group(iter(node.children), parent, _font)
            elif token in commands.GLOBAL_FONTS.keys():
                _font = commands.GLOBAL_FONTS.get(token)
            elif node.children is None:
                self._convert_symbol(node, parent, _font)
            else:
                attributes = node.attributes or {}
                _row = SubElement(parent, "mrow", attrib=attributes)
                self._convert_group(iter(node.children), _row, _font)

    def _convert_command(self, node: Node, parent: Element, font: Optional[dict[str, Optional[str]]] = None) -> None:
        command = node.token
        modifier = node.modifier

        if command in (commands.SUBSTACK, commands.SMALLMATRIX):
            parent = SubElement(parent, "mstyle", scriptlevel="1")
        elif command == commands.CASES:
            parent = SubElement(parent, "mrow")
            lbrace = SubElement(parent, "mo", {"stretchy": "true", "fence": "true", "form": "prefix"})
            lbrace.text = "&#x{};".format(convert_symbol(commands.LBRACE))
        elif command in (commands.DBINOM, commands.DFRAC):
            parent = SubElement(parent, "mstyle", displaystyle="true", scriptlevel="0")
        elif command == commands.HPHANTOM:
            parent = SubElement(parent, "mpadded", height="0", depth="0")
        elif command == commands.VPHANTOM:
            parent = SubElement(parent, "mpadded", width="0")
        elif command in (commands.TBINOM, commands.HBOX, commands.MBOX, commands.TFRAC):
            parent = SubElement(parent, "mstyle", displaystyle="false", scriptlevel="0")
        elif command in (commands.MOD, commands.PMOD, commands.POD):
            SubElement(parent, "mspace", width="1em")

        tag, attributes = copy.deepcopy(commands.CONVERSION_MAP[command])

        if node.attributes is not None and node.token != commands.SKEW:
            attributes.update(node.attributes)

        if command == commands.LEFT:
            parent = SubElement(parent, "mrow")

        self._append_delimiter_element(node, parent, is_prefix=True)

        alignment, column_lines = self._get_alignment_and_column_lines(node.alignment)

        if column_lines:
            attributes["columnlines"] = column_lines

        if command == commands.SUBSUP and node.children is not None and node.children[0].token == commands.GCD:
            tag = "munderover"
        elif command == commands.SUPERSCRIPT and modifier in (commands.LIMITS, commands.OVERBRACE):
            tag = "mover"
        elif command == commands.SUBSCRIPT and modifier in (commands.LIMITS, commands.UNDERBRACE):
            tag = "munder"
        elif command == commands.SUBSUP and modifier in (commands.LIMITS, commands.OVERBRACE, commands.UNDERBRACE):
            tag = "munderover"
        elif command in commands.EXTENSIBLE_ARROWS and node.children is not None and len(node.children) == 2:
            tag = "munderover"

        element = SubElement(parent, tag, attributes)

        if command in commands.LIMIT:
            element.text = command[1:]
        elif command in (commands.MOD, commands.PMOD):
            element.text = "mod"
            SubElement(parent, "mspace", width="0.333em")
        elif command == commands.POD:
            pass
        elif command == commands.BMOD:
            element.text = "mod"
        elif command in commands.EXTENSIBLE_ARROWS:
            style = SubElement(element, "mstyle", scriptlevel="0")
            arrow = SubElement(style, "mo")
            arrow.text = commands.EXTENSIBLE_ARROWS[command]
        elif command == commands.BRA:
            SubElement(element, "mo", stretchy="false").text = "&#x27E8;"
        elif command == commands.KET:
            SubElement(element, "mo").text = "&#x2223;"
        elif command == commands.BRAKET:
            SubElement(element, "mo", stretchy="false").text = "&#x27E8;"
        elif node.text is not None:
            if command == commands.MIDDLE:
                element.text = "&#x{};".format(convert_symbol(node.text))
            elif command == commands.HBOX:
                mtext: Optional[Element] = element
                for text, mode in self._separate_by_mode(node.text):
                    if mode == Mode.TEXT:
                        if mtext is None:
                            mtext = SubElement(parent, tag, attributes)
                        mtext.text = text.replace(" ", "&#x000A0;")
                        self._set_font(mtext, "mtext", font)
                        mtext = None
                    else:
                        _row = SubElement(parent, "mrow")
                        self._convert_group(iter(walk(text, macros=self.macros)), _row)
            else:
                if command in (
                    commands.FBOX,
                    commands.LLAP,
                    commands.RLAP,
                    commands.CLAP,
                    commands.COLORBOX,
                    commands.FCOLORBOX,
                ):
                    element = SubElement(element, "mtext")
                if command == commands.TAG:
                    element.text = f"({node.text})"
                elif command == commands.TAGSTAR:
                    element.text = node.text
                else:
                    element.text = node.text.replace(" ", "&#x000A0;")
                self._set_font(element, "mtext", font)
        elif node.delimiter is not None and command not in (commands.FRAC, commands.GENFRAC):
            if node.delimiter != ".":
                symbol = convert_symbol(node.delimiter)
                element.text = node.delimiter if symbol is None else "&#x{};".format(symbol)

        if node.children is not None:
            _parent = element
            if command in (commands.LEFT, commands.MOD, commands.PMOD, commands.POD):
                _parent = parent
            if command in commands.MATRICES:
                if command == commands.CASES:
                    alignment = "l"
                elif command in (commands.SPLIT, commands.ALIGN, commands.ALIGNSTAR):
                    alignment = "rl"
                self._convert_matrix(iter(node.children), _parent, command, alignment=alignment)
            elif command == commands.CFRAC:
                for child in node.children:
                    p = SubElement(_parent, "mstyle", displaystyle="false", scriptlevel="0")
                    self._convert_group(iter([child]), p, font)
            elif command == commands.SIDESET:
                left, right = node.children
                self._convert_group(iter([left]), _parent, font)
                fill = SubElement(_parent, "mstyle", scriptlevel="0")
                SubElement(fill, "mspace", width="-0.167em")
                self._convert_group(iter([right]), _parent, font)
            elif command == commands.SKEW:
                child = node.children[0]
                new_node = Node(
                    token=child.token,
                    children=(
                        Node(
                            token=commands.BRACES,
                            children=(*child.children, Node(token=commands.MKERN, attributes=node.attributes)),
                        ),
                    ),
                )
                self._convert_group(iter([new_node]), _parent, font)
            elif command in commands.EXTENSIBLE_ARROWS:
                for child in node.children:
                    padded = SubElement(
                        _parent,
                        "mpadded",
                        {"width": "+0.833em", "lspace": "0.556em", "voffset": "-.2em", "height": "-.2em"},
                    )
                    self._convert_group(iter([child]), padded, font)
                    SubElement(padded, "mspace", depth=".25em")
            else:
                self._convert_group(iter(node.children), _parent, font)

        if command in commands.DIACRITICS:
            text, diacritic_attrs = copy.deepcopy(commands.DIACRITICS[command])
            SubElement(element, "mo", diacritic_attrs).text = text

        if command == commands.BRA:
            SubElement(element, "mo").text = "&#x2223;"
        elif command == commands.KET:
            SubElement(element, "mo", stretchy="false").text = "&#x27E9;"
        elif command == commands.BRAKET:
            SubElement(element, "mo", stretchy="false").text = "&#x27E9;"

        self._append_delimiter_element(node, parent, is_prefix=False)

    def _append_delimiter_element(self, node: Node, parent: Element, is_prefix: bool) -> None:
        delimiter_index = 0 if is_prefix else 1
        size = "2.047em"
        if parent.attrib.get("displaystyle") == "false" or node.token == commands.TBINOM:
            size = "1.2em"
        if node.token in (r"\pmatrix", commands.PMOD, commands.POD):
            self._convert_and_append_command(r"\lparen" if is_prefix else r"\rparen", parent)
        elif node.token in (commands.BINOM, commands.DBINOM, commands.TBINOM):
            self._convert_and_append_command(
                r"\lparen" if is_prefix else r"\rparen", parent, {"minsize": size, "maxsize": size}
            )
        elif node.token == r"\bmatrix":
            self._convert_and_append_command(r"\lbrack" if is_prefix else r"\rbrack", parent)
        elif node.token == r"\Bmatrix":
            self._convert_and_append_command(r"\lbrace" if is_prefix else r"\rbrace", parent)
        elif node.token == r"\vmatrix":
            self._convert_and_append_command(r"\vert", parent)
        elif node.token == r"\Vmatrix":
            self._convert_and_append_command(r"\Vert", parent)
        elif (
            node.token in (commands.FRAC, commands.GENFRAC)
            and node.delimiter is not None
            and node.delimiter[delimiter_index] != "."
        ):
            self._convert_and_append_command(
                node.delimiter[delimiter_index], parent, {"minsize": size, "maxsize": size}
            )
        elif not is_prefix and node.token == commands.SKEW and node.attributes is not None:
            SubElement(parent, "mspace", width="-" + node.attributes["width"])

    def _convert_symbol(self, node: Node, parent: Element, font: Optional[dict[str, Optional[str]]] = None) -> None:
        token = node.token
        attributes = node.attributes or {}
        symbol = convert_symbol(token)
        if token == MULTIPRIMES:
            count = int(node.text or "0")
            element = SubElement(parent, "mi", attrib=attributes)
            element.text = "&#x02032;" * count
            return
        if NUMBER_PATTERN.match(token):
            element = SubElement(parent, "mn", attrib=attributes)
            element.text = token
            self._set_font(element, element.tag, font)
        elif token in OPERATORS:
            element = SubElement(parent, "mo", attrib=attributes)
            element.text = token if symbol is None else "&#x{};".format(symbol)
            if token == r"\|":
                element.attrib["fence"] = "false"
            if token == r"\smallint":
                element.attrib["largeop"] = "false"
            if token in ("(", ")", "[", "]", "|", r"\|", r"\{", r"\}", r"\surd"):
                element.attrib["stretchy"] = "false"
                self._set_font(element, "fence", font)
            else:
                self._set_font(element, element.tag, font)
        elif (
            symbol
            and (
                int(symbol, 16) in range(int("2200", 16), int("22FF", 16) + 1)
                or int(symbol, 16) in range(int("2190", 16), int("21FF", 16) + 1)
            )
            or symbol == "."
        ):
            element = SubElement(parent, "mo", attrib=attributes)
            element.text = "&#x{};".format(symbol)
            self._set_font(element, element.tag, font)
        elif token in (r"\ ", "~", commands.NOBREAKSPACE, commands.SPACE):
            element = SubElement(parent, "mtext", attrib=attributes)
            element.text = "&#x000A0;"
            self._set_font(element, "mtext", font)
        elif token == commands.NOT:
            mpadded = SubElement(parent, "mpadded", width="0")
            element = SubElement(mpadded, "mtext")
            element.text = "&#x029F8;"
        elif token in MOVABLE_LIMIT_TEXTS:
            element = SubElement(parent, "mo", attrib={"movablelimits": "true", **attributes})
            element.text = MOVABLE_LIMIT_TEXTS[token]
            self._set_font(element, element.tag, font)
        elif token in (commands.MATHSTRUT, commands.STRUT):
            mpadded = SubElement(parent, "mpadded", width="0px")
            mphantom = SubElement(mpadded, "mphantom")
            SubElement(mphantom, "mo", stretchy="false").text = "&#x00028;"
        elif token == commands.IDOTSINT:
            _parent = SubElement(parent, "mrow", attrib=attributes)
            for s in ("&#x0222B;", "&#x022EF;", "&#x0222B;"):
                element = SubElement(_parent, "mo")
                element.text = s
        elif token in (commands.LATEX, commands.TEX):
            _parent = SubElement(parent, "mrow", attrib=attributes)
            if token == commands.LATEX:
                mi_l = SubElement(_parent, "mi")
                mi_l.text = "L"
                SubElement(_parent, "mspace", width="-.325em")
                mpadded = SubElement(_parent, "mpadded", height="+.21ex", depth="-.21ex", voffset="+.21ex")
                mstyle = SubElement(mpadded, "mstyle", displaystyle="false", scriptlevel="1")
                mrow = SubElement(mstyle, "mrow")
                mi_a = SubElement(mrow, "mi")
                mi_a.text = "A"
                SubElement(_parent, "mspace", width="-.17em")
                self._set_font(mi_l, mi_l.tag, font)
                self._set_font(mi_a, mi_a.tag, font)
            mi_t = SubElement(_parent, "mi")
            mi_t.text = "T"
            SubElement(_parent, "mspace", width="-.14em")
            mpadded = SubElement(_parent, "mpadded", height="-.5ex", depth="+.5ex", voffset="-.5ex")
            mrow = SubElement(mpadded, "mrow")
            mi_e = SubElement(mrow, "mi")
            mi_e.text = "E"
            SubElement(_parent, "mspace", width="-.115em")
            mi_x = SubElement(_parent, "mi")
            mi_x.text = "X"
            self._set_font(mi_t, mi_t.tag, font)
            self._set_font(mi_e, mi_e.tag, font)
            self._set_font(mi_x, mi_x.tag, font)
        elif token.startswith(commands.OPERATORNAME):
            for prefix in (commands.OPERATORNAMEWITHLIMITS, commands.OPERATORNAMESTAR, commands.OPERATORNAME):
                if token.startswith(prefix):
                    attrib = {"movablelimits": "true", **attributes} if prefix != commands.OPERATORNAME else attributes
                    element = SubElement(parent, "mo", attrib=attrib)
                    element.text = token[len(prefix) + 1 : -1]
                    break
        elif token.startswith(commands.BACKSLASH):
            element = SubElement(parent, "mi", attrib=attributes)
            if symbol:
                element.text = "&#x{};".format(symbol)
            elif token in commands.FUNCTIONS:
                element.text = token[1:]
            else:
                element.text = token
            self._set_font(element, element.tag, font)
        else:
            element = SubElement(parent, "mi", attrib=attributes)
            element.text = token
            self._set_font(element, element.tag, font)

    @staticmethod
    def _set_font(element: Element, key: str, font: Optional[dict[str, Optional[str]]]) -> None:
        if font is None:
            return
        _font = font[key]
        if _font is not None:
            element.attrib["mathvariant"] = _font

    @staticmethod
    def _set_cell_alignment(cell: Element, hfil_indexes: list[bool]) -> None:
        if cell is not None and any(hfil_indexes) and len(hfil_indexes) > 1:
            if hfil_indexes[0] and not hfil_indexes[-1]:
                cell.attrib["columnalign"] = "right"
            elif not hfil_indexes[0] and hfil_indexes[-1]:
                cell.attrib["columnalign"] = "left"

    @staticmethod
    def _get_column_alignment(
        alignment: Optional[str], column_alignment: Optional[str], column_index: int
    ) -> tuple[Optional[str], int]:
        if alignment:
            try:
                column_alignment = COLUMN_ALIGNMENT_MAP.get(alignment[column_index])
            except IndexError:
                column_alignment = COLUMN_ALIGNMENT_MAP.get(alignment[column_index % len(alignment)])
            column_index += 1
        return column_alignment, column_index

    @staticmethod
    def _make_matrix_cell(row: Element, column_alignment: Optional[str]) -> Element:
        if column_alignment:
            return SubElement(row, "mtd", columnalign=column_alignment)
        return SubElement(row, "mtd")

    @staticmethod
    def _get_alignment_and_column_lines(
        alignment: Optional[str] = None,
    ) -> tuple[Optional[str], Optional[str]]:
        if alignment is None:
            return None, None
        if "|" not in alignment:
            return alignment, None
        _alignment = ""
        column_lines: list[str] = []
        for c in alignment:
            if c == "|":
                column_lines.append("solid")
            else:
                _alignment += c
            if len(_alignment) - len(column_lines) == 2:
                column_lines.append("none")
        return _alignment, " ".join(column_lines)

    @staticmethod
    def _separate_by_mode(text: str) -> Iterator[tuple[str, Mode]]:
        string = ""
        is_math_mode = False
        for match in MATH_MODE_PATTERN.findall(text):
            if match == "$":
                yield string, Mode.MATH if is_math_mode else Mode.TEXT
                string = ""
                is_math_mode = not is_math_mode
            else:
                string += match
        if string:
            yield string, Mode.MATH if is_math_mode else Mode.TEXT

    @staticmethod
    def _convert_and_append_command(command: str, parent: Element, attributes: Optional[dict[str, str]] = None) -> None:
        code_point = convert_symbol(command)
        mo = SubElement(parent, "mo", attributes if attributes is not None else {})
        mo.text = "&#x{};".format(code_point) if code_point else command


def convert(
    latex: str,
    xmlns: str = "http://www.w3.org/1998/Math/MathML",
    display: str = "inline",
    parent: Optional[Element] = None,
) -> str:
    return Converter(xmlns=xmlns, display=display).convert(latex, parent=parent)


def convert_to_element(
    latex: str,
    xmlns: str = "http://www.w3.org/1998/Math/MathML",
    display: str = "inline",
    parent: Optional[Element] = None,
) -> Element:
    return Converter(xmlns=xmlns, display=display).convert_to_element(latex, parent=parent)


def main() -> None:  # pragma: no cover
    import argparse
    import sys

    parser = argparse.ArgumentParser(description="Pure Python library for LaTeX to MathML conversion")
    parser.add_argument("-V", "--version", dest="version", action="store_true", required=False, help="Show version")
    parser.add_argument("-b", "--block", dest="block", action="store_true", required=False, help="Display block")

    required = parser.add_argument_group("required arguments")

    group = required.add_mutually_exclusive_group(required=False)
    group.add_argument("-t", "--text", dest="text", type=str, required=False, help="Text")
    group.add_argument("-f", "--file", dest="file", type=str, required=False, help="File")
    group.add_argument("-s", "--stdin", dest="stdin", action="store_true", required=False, help="Stdin")

    arguments = parser.parse_args()
    display = "block" if arguments.block else "inline"

    if arguments.version:
        import latex2mathml

        print("latex2mathml", latex2mathml.__version__)
    elif arguments.text:
        print(convert(arguments.text, display=display))
    elif arguments.file:
        with open(arguments.file) as f:
            print(convert(f.read(), display=display))
    elif arguments.stdin:
        print(convert(sys.stdin.read(), display=display))


if __name__ == "__main__":  # pragma: no cover
    main()



from itertools import chain
from typing import Any, Iterator, NamedTuple, Optional

from latex2mathml import commands
from latex2mathml.exceptions import (
    DenominatorNotFoundError,
    DoubleSubscriptsError,
    DoubleSuperscriptsError,
    ExtraLeftOrMissingRightError,
    InvalidAlignmentError,
    InvalidStyleForGenfracError,
    InvalidWidthError,
    LimitsMustFollowMathOperatorError,
    MissingEndError,
    MissingSuperScriptOrSubscriptError,
    NoAvailableTokensError,
    NumeratorNotFoundError,
)
from latex2mathml.symbols_parser import convert_symbol
from latex2mathml.tokenizer import tokenize

MULTIPRIMES = "multiprimes"
MAX_MACRO_DEPTH = 100


class Node(NamedTuple):
    token: str
    children: Optional[tuple[Any, ...]] = None
    delimiter: Optional[str] = None
    alignment: Optional[str] = None
    text: Optional[str] = None
    attributes: Optional[dict[str, str]] = None
    modifier: Optional[str] = None


def walk(data: str, display: str = "inline", macros: Optional[dict[str, tuple[list[str], int]]] = None) -> list[Node]:
    tokens = tokenize(data)
    block = display == "block"
    return _walk(tokens, block=block, macros={} if macros is None else macros)


def _walk(
    tokens: Iterator[str],
    terminator: Optional[str] = None,
    limit: int = 0,
    block: bool = False,
    macros: Optional[dict[str, tuple[list[str], int]]] = None,
    depth: int = 0,
) -> list[Node]:
    _macros = {} if macros is None else macros
    group: list[Node] = []
    token: str
    has_available_tokens = False
    for token in tokens:
        has_available_tokens = True
        if token == terminator:
            delimiter = None
            if terminator == commands.RIGHT:
                delimiter = next(tokens)
            group.append(Node(token=token, delimiter=delimiter))
            break
        elif (token == commands.RIGHT != terminator) or (token == commands.MIDDLE and terminator != commands.RIGHT):
            raise ExtraLeftOrMissingRightError
        elif token == commands.LEFT:
            delimiter = next(tokens)
            children = tuple(
                _walk(tokens, terminator=commands.RIGHT, macros=_macros)
            )  # make \right as a child of \left
            if len(children) == 0 or children[-1].token != commands.RIGHT:
                raise ExtraLeftOrMissingRightError
            node = Node(token=token, children=children, delimiter=delimiter)
        elif token == commands.OPENING_BRACE:
            children = tuple(_walk(tokens, terminator=commands.CLOSING_BRACE, macros=_macros))
            if len(children) and children[-1].token == commands.CLOSING_BRACE:
                children = children[:-1]
            node = Node(token=commands.BRACES, children=children)
        elif token in (commands.SUBSCRIPT, commands.SUPERSCRIPT):
            try:
                previous = group.pop()
            except IndexError:
                previous = Node(token="")  # left operand can be empty if not present

            if token == previous.token == commands.SUBSCRIPT:
                raise DoubleSubscriptsError
            if (token == previous.token == commands.SUPERSCRIPT) and (
                previous.children is not None
                and len(previous.children) >= 2
                and previous.children[1].token != commands.PRIME
            ):
                raise DoubleSuperscriptsError

            modifier = None
            if previous.token in (commands.LIMITS, commands.NOLIMITS):
                modifier = previous.token
                try:
                    previous = group.pop()
                    if not previous.token.startswith("\\"):  # TODO: Complete list of operators
                        raise LimitsMustFollowMathOperatorError
                except IndexError:
                    raise LimitsMustFollowMathOperatorError
            elif block and previous.token in (commands.SUMMATION, commands.PRODUCT):
                # block summation and product should result in limited sub/sup
                modifier = commands.LIMITS

            if token == commands.SUBSCRIPT and previous.token == commands.SUPERSCRIPT and previous.children is not None:
                children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
                node = Node(
                    token=commands.SUBSUP,
                    children=(previous.children[0], *children, previous.children[1]),
                    modifier=previous.modifier,
                )
            elif (
                token == commands.SUPERSCRIPT and previous.token == commands.SUBSCRIPT and previous.children is not None
            ):
                children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
                node = Node(token=commands.SUBSUP, children=(*previous.children, *children), modifier=previous.modifier)
            elif (
                token == commands.SUPERSCRIPT
                and previous.token == commands.SUPERSCRIPT
                and previous.children is not None
                and previous.children[1].token == commands.PRIME
            ):
                children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))

                node = Node(
                    token=commands.SUPERSCRIPT,
                    children=(
                        previous.children[0],
                        Node(token=commands.BRACES, children=(previous.children[1], *children)),
                    ),
                    modifier=previous.modifier,
                )
            else:
                try:
                    children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
                except NoAvailableTokensError:
                    raise MissingSuperScriptOrSubscriptError
                if previous.token in (commands.OVERBRACE, commands.UNDERBRACE):
                    modifier = previous.token
                node = Node(token=token, children=(previous, *children), modifier=modifier)
        elif token == commands.APOSTROPHE:
            try:
                previous = group.pop()
            except IndexError:
                previous = Node(token="")  # left operand can be empty if not present

            prev_is_super_with_children = (
                previous.token == commands.SUPERSCRIPT and previous.children is not None and len(previous.children) >= 2
            )
            prev_prime = previous.children[1] if prev_is_super_with_children and previous.children else None
            prev_is_prime_token = prev_prime is not None and (
                prev_prime.token in commands.PRIME_UPGRADE
                or prev_prime.token == commands.QPRIME
                or prev_prime.token == MULTIPRIMES
            )

            if prev_is_super_with_children and not prev_is_prime_token:
                raise DoubleSuperscriptsError

            if prev_is_prime_token and previous.children is not None and prev_prime is not None:
                if prev_prime.token in commands.PRIME_UPGRADE:
                    new_prime = Node(token=commands.PRIME_UPGRADE[prev_prime.token])
                elif prev_prime.token == commands.QPRIME:
                    new_prime = Node(token=MULTIPRIMES, text="5")
                else:
                    new_prime = Node(token=MULTIPRIMES, text=str(int(prev_prime.text or "0") + 1))
                node = Node(token=commands.SUPERSCRIPT, children=(previous.children[0], new_prime))
            elif previous.token == commands.SUBSCRIPT and previous.children is not None:
                node = Node(
                    token=commands.SUBSUP,
                    children=(*previous.children, Node(token=commands.PRIME)),
                    modifier=previous.modifier,
                )
            else:
                node = Node(token=commands.SUPERSCRIPT, children=(previous, Node(token=commands.PRIME)))
        elif token in commands.COMMANDS_WITH_TWO_PARAMETERS:
            children = tuple(_walk(tokens, terminator=terminator, limit=2, macros=_macros))
            if token in (commands.OVERSET, commands.STACKREL, commands.UNDERSET):
                children = children[::-1]
            node = Node(token=token, children=children)
        elif token in commands.COMMANDS_WITH_ONE_PARAMETER or (
            token.startswith(commands.MATH) and token not in commands.MATH_NON_FONT_COMMANDS
        ):
            children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            node = Node(token=token, children=children)
        elif token == commands.NOT:
            try:
                next_node = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))[0]
                if next_node.token.startswith("\\"):
                    negated_symbol = r"\n" + next_node.token[1:]
                    symbol = convert_symbol(negated_symbol)
                    if symbol:
                        node = Node(token=negated_symbol)
                        group.append(node)
                        continue
                node = Node(token=token)
                group.extend((node, next_node))
                continue
            except NoAvailableTokensError:
                node = Node(token=token)
        elif token in commands.EXTENSIBLE_ARROWS:
            children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            if children[0].token == commands.OPENING_BRACKET:
                children = (
                    Node(
                        token=commands.BRACES,
                        children=tuple(_walk(tokens, terminator=commands.CLOSING_BRACKET, macros=_macros))[:-1],
                    ),
                    *tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros)),
                )
            node = Node(token=token, children=children)
        elif token in (commands.HSKIP, commands.HSPACE, commands.KERN, commands.MKERN, commands.MSKIP, commands.MSPACE):
            children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            node = Node(token=token, attributes={"width": _unwrap_token(children[0])})
        elif token in (commands.RAISE, commands.LOWER, commands.MOVELEFT, commands.MOVERIGHT):
            dim_children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            dim = _unwrap_token(dim_children[0]) if dim_children else "0"
            children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            if token == commands.RAISE:
                attributes = {"voffset": dim, "height": f"+{dim}", "depth": f"-{dim}"}
            elif token == commands.LOWER:
                attributes = {"voffset": f"-{dim}", "height": f"-{dim}", "depth": f"+{dim}"}
            elif token == commands.MOVELEFT:
                attributes = {"lspace": f"-{dim}"}
            else:
                attributes = {"lspace": dim}
            node = Node(token=token, children=children, attributes=attributes)
        elif token == commands.RULE:
            dims = []
            for _ in range(2):
                arg = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))[0]
                dims.append(_unwrap_token(arg))
            node = Node(token=token, attributes={"width": dims[0], "height": dims[1]})
        elif token == commands.SMASH:
            children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            attributes = {"height": "0px", "depth": "0px"}
            if children[0].token == commands.OPENING_BRACKET:
                opt = tuple(_walk(tokens, terminator=commands.CLOSING_BRACKET, macros=_macros))[:-1]
                if opt and opt[0].token == "b":
                    attributes = {"depth": "0px"}
                elif opt and opt[0].token == "t":
                    attributes = {"height": "0px"}
                children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            node = Node(token=token, children=children, attributes=attributes)
        elif token == commands.TEXTCOLOR:
            color = next(tokens)
            children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            node = Node(token=commands.COLOR, children=children, attributes={"mathcolor": color})
        elif token in (commands.COLORBOX, commands.FCOLORBOX):
            arg_count = 3 if token == commands.FCOLORBOX else 2
            args = []
            for _ in range(arg_count):
                arg_node = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))[0]
                args.append("".join(c.token for c in arg_node.children) if arg_node.children else "")
            if token == commands.FCOLORBOX:
                attributes = {"mathbackground": args[1], "border-color": args[0]}
            else:
                attributes = {"mathbackground": args[0]}
            node = Node(token=token, text=args[-1], attributes=attributes)
        elif token == commands.COLOR:
            attributes = {"mathcolor": next(tokens)}
            children = tuple(_walk(tokens, terminator=terminator, macros=_macros))
            sibling = None
            if len(children) and children[-1].token == terminator:
                children, sibling = children[:-1], children[-1]
            group.append(Node(token=token, children=children, attributes=attributes))
            if sibling:
                group.append(sibling)
            break
        elif token in (commands.LEFTROOT, commands.UPROOT):
            # Consume the numeric argument but discard — MathML has no root index positioning
            tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            continue
        elif token == commands.MATHCHOICE:
            choices = tuple(_walk(tokens, terminator=terminator, limit=4, macros=_macros))
            # In block (display) mode use arg 0, in inline (text) mode use arg 1
            choice = choices[0] if block else choices[1]
            if choice.children:
                for child in choice.children:
                    group.append(child)
            else:
                group.append(choice)
            continue
        elif token in (commands.CLASS, commands.STYLE):
            attr_name = "class" if token == commands.CLASS else "style"
            attributes = {attr_name: next(tokens)}
            next_node = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))[0]
            node = next_node._replace(attributes=attributes)
        elif token in (
            *commands.BIG.keys(),
            *commands.BIG_OPEN_CLOSE.keys(),
            commands.CLAP,
            commands.EMPH,
            commands.FBOX,
            commands.HBOX,
            commands.LLAP,
            commands.MBOX,
            commands.MIDDLE,
            commands.RLAP,
            commands.TAG,
            commands.TAGSTAR,
            commands.TEXT,
            commands.TEXTBF,
            commands.TEXTIT,
            commands.TEXTMD,
            commands.TEXTNORMAL,
            commands.TEXTRM,
            commands.TEXTSF,
            commands.TEXTTT,
            commands.TEXTUP,
            commands.VERB,
        ):
            node = Node(token=token, text=next(tokens))
        elif token == commands.HREF:
            attributes = {"href": next(tokens)}
            children = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))
            node = Node(token=token, children=children, attributes=attributes)
        elif token in (
            commands.ABOVE,
            commands.ATOP,
            commands.ABOVEWITHDELIMS,
            commands.ATOPWITHDELIMS,
            commands.BRACE,
            commands.BRACK,
            commands.CHOOSE,
            commands.OVER,
        ):
            attributes = None
            delimiter = None

            if token == commands.ABOVEWITHDELIMS:
                delimiter = next(tokens).lstrip("\\") + next(tokens).lstrip("\\")
            elif token == commands.ATOPWITHDELIMS:
                attributes = {"linethickness": "0"}
                delimiter = next(tokens).lstrip("\\") + next(tokens).lstrip("\\")
            elif token == commands.BRACE:
                delimiter = "{}"
            elif token == commands.BRACK:
                delimiter = "[]"
            elif token == commands.CHOOSE:
                delimiter = "()"

            if token in (commands.ABOVE, commands.ABOVEWITHDELIMS):
                dimension_node = tuple(_walk(tokens, terminator=terminator, limit=1, macros=_macros))[0]
                dimension = _unwrap_token(dimension_node)
                attributes = {"linethickness": dimension}
            elif token in (commands.ATOP, commands.BRACE, commands.BRACK, commands.CHOOSE):
                attributes = {"linethickness": "0"}

            denominator = tuple(_walk(tokens, terminator=terminator, macros=_macros))

            sibling = None
            if len(denominator) and denominator[-1].token == terminator:
                denominator, sibling = denominator[:-1], denominator[-1]

            if len(denominator) == 0:
                if token in (commands.BRACE, commands.BRACK):
                    denominator = (Node(token=commands.BRACES, children=()),)
                else:
                    raise DenominatorNotFoundError
            if len(group) == 0:
                if token in (commands.BRACE, commands.BRACK):
                    group = [Node(token=commands.BRACES, children=())]
                else:
                    raise NumeratorNotFoundError
            if len(denominator) > 1:
                denominator = (Node(token=commands.BRACES, children=denominator),)

            if len(group) == 1:
                children = (*group, *denominator)
            else:
                children = (Node(token=commands.BRACES, children=tuple(group)), *denominator)
            group = [Node(token=commands.FRAC, children=children, attributes=attributes, delimiter=delimiter)]
            if sibling is not None:
                group.append(sibling)
            break
        elif token == commands.SQRT:
            root_nodes = None
            next_node = tuple(_walk(tokens, limit=1, macros=_macros))[0]
            if next_node.token == commands.OPENING_BRACKET:
                root_nodes = tuple(_walk(tokens, terminator=commands.CLOSING_BRACKET, macros=_macros))[:-1]
                next_node = tuple(_walk(tokens, limit=1, macros=_macros))[0]
                if len(root_nodes) > 1:
                    root_nodes = (Node(token=commands.BRACES, children=root_nodes),)

            if root_nodes:
                node = Node(token=commands.ROOT, children=(next_node, *root_nodes))
            else:
                node = Node(token=token, children=(next_node,))
        elif token == commands.ROOT:
            root_nodes = tuple(_walk(tokens, terminator=r"\of", macros=_macros))[:-1]
            next_node = tuple(_walk(tokens, limit=1, macros=_macros))[0]
            if len(root_nodes) > 1:
                root_nodes = (Node(token=commands.BRACES, children=root_nodes),)
            if root_nodes:
                node = Node(token=token, children=(next_node, *root_nodes))
            else:
                node = Node(token=token, children=(next_node, Node(token=commands.BRACES, children=())))
        elif token in commands.MATRICES:
            children = tuple(_walk(tokens, terminator=terminator, macros=_macros))
            sibling = None
            if len(children) and children[-1].token == terminator:
                children, sibling = children[:-1], children[-1]
            if len(children) == 1 and children[0].token == commands.BRACES and children[0].children:
                children = children[0].children
            if sibling is not None:
                group.extend([Node(token=token, children=children, alignment=""), sibling])
                break
            else:
                node = Node(token=token, children=children, alignment="")
        elif token == commands.GENFRAC:
            delimiter = next(tokens).lstrip("\\") + next(tokens).lstrip("\\")
            dimension_node, style_node = tuple(_walk(tokens, terminator=terminator, limit=2, macros=_macros))
            dimension = _unwrap_token(dimension_node)
            style = _get_style(style_node)
            attributes = {"linethickness": dimension}
            children = tuple(_walk(tokens, terminator=terminator, limit=2, macros=_macros))
            group.extend(
                [Node(token=style), Node(token=token, children=children, delimiter=delimiter, attributes=attributes)]
            )
            break
        elif token == commands.SIDESET:
            left, right, operator = tuple(_walk(tokens, terminator=terminator, limit=3, macros=_macros))
            left_token, left_children = _make_subsup(left)
            right_token, right_children = _make_subsup(right)
            attributes = {"movablelimits": "false"}
            node = Node(
                token=token,
                children=(
                    Node(
                        token=left_token,
                        children=(
                            Node(
                                token=commands.VPHANTOM,
                                children=(
                                    Node(token=operator.token, children=operator.children, attributes=attributes),
                                ),
                            ),
                            *left_children,
                        ),
                    ),
                    Node(
                        token=right_token,
                        children=(
                            Node(token=operator.token, children=operator.children, attributes=attributes),
                            *right_children,
                        ),
                    ),
                ),
            )
        elif token == commands.SKEW:
            width_node, child = tuple(_walk(tokens, terminator=terminator, limit=2, macros=_macros))
            width = width_node.token
            if width == commands.BRACES:
                if width_node.children is None or len(width_node.children) == 0:
                    raise InvalidWidthError
                width = width_node.children[0].token
            if not width.isdigit():
                raise InvalidWidthError
            node = Node(token=token, children=(child,), attributes={"width": f"{0.0555 * int(width):.3f}em"})
        elif token.startswith(commands.BEGIN):
            node = _get_environment_node(token, tokens, macros=_macros, block=block)
        elif token == commands.NEWCOMMAND:
            _parse_newcommand(tokens, _macros)
            continue
        elif token == commands.DEF:
            _parse_def(tokens, _macros)
            continue
        elif token == commands.DECLAREMATHOPERATOR:
            _parse_declare_math_operator(tokens, _macros)
            continue
        elif token == commands.NEWENVIRONMENT:
            _parse_newenvironment(tokens, _macros)
            continue
        elif token in _macros:
            if depth >= MAX_MACRO_DEPTH:
                raise RecursionError(f"Maximum macro expansion depth ({MAX_MACRO_DEPTH}) exceeded")
            expanded_tokens = _expand_macro(token, tokens, _macros)
            if not expanded_tokens:
                continue
            chained = chain(iter(expanded_tokens), tokens)
            remaining_limit = max(0, limit - len(group)) if limit else 0
            group.extend(
                _walk(
                    chained, terminator=terminator, block=block, macros=_macros, depth=depth + 1, limit=remaining_limit
                )
            )
            break
        else:
            node = Node(token=token)

        group.append(node)

        if limit and len(group) >= limit:
            break
    if not has_available_tokens:
        raise NoAvailableTokensError
    return group


def _make_subsup(node: Node) -> tuple[str, tuple[Node, ...]]:
    if node.token != commands.BRACES:
        raise MissingSuperScriptOrSubscriptError
    if (
        node.children is not None
        and len(node.children) > 0
        and node.children[0].children is not None
        and 2 <= len(node.children[0].children) <= 3
        and node.children[0].token
        in (
            commands.SUBSUP,
            commands.SUBSCRIPT,
            commands.SUPERSCRIPT,
        )
    ):
        token = node.children[0].token
        children = node.children[0].children[1:]
        return token, children
    return "", ()


def _unwrap_token(node: Node) -> str:
    dimension = node.token
    if node.token == commands.BRACES and node.children is not None:
        dimension = node.children[0].token
    return dimension


def _get_style(node: Node) -> str:
    style = _unwrap_token(node)
    if style == "0":
        return commands.DISPLAYSTYLE
    if style == "1":
        return commands.TEXTSTYLE
    if style == "2":
        return commands.SCRIPTSTYLE
    if style == "3":
        return commands.SCRIPTSCRIPTSTYLE
    raise InvalidStyleForGenfracError


def _get_environment_node(
    token: str,
    tokens: Iterator[str],
    macros: Optional[dict[str, tuple[list[str], int]]] = None,
    block: bool = False,
) -> Node:
    _macros = {} if macros is None else macros
    start_index = token.index("{") + 1
    environment = token[start_index:-1]
    env_key = f"\\begin{{{environment}}}"
    if env_key in _macros:
        begin_body, nargs = _macros[env_key]
        end_body, _ = _macros.get(f"\\end{{{environment}}}", ([], 0))
        args: list[list[str]] = []
        for _ in range(nargs):
            args.append(_consume_brace_arg(tokens))
        terminator = rf"{commands.END}{{{environment}}}"
        raw_tokens: list[str] = []
        found_end = False
        for t in tokens:
            if t == terminator:
                found_end = True
                break
            raw_tokens.append(t)
        if not found_end:
            raise MissingEndError
        begin_expanded = _substitute_params(begin_body, args)
        end_expanded = _substitute_params(end_body, args)
        expanded = [*begin_expanded, *raw_tokens, *end_expanded]
        result = _walk(iter(expanded), macros=_macros, block=block)
        if len(result) == 1:
            return result[0]
        return Node(token=commands.BRACES, children=tuple(result))
    terminator = rf"{commands.END}{{{environment}}}"
    children = tuple(_walk(tokens, terminator=terminator, macros=macros, block=block))
    if len(children) and children[-1].token != terminator:
        raise MissingEndError
    children = children[:-1]
    alignment = ""

    if len(children) and children[0].token == commands.OPENING_BRACKET:
        children_iter = iter(children)
        next(children_iter)  # remove BRACKET
        for c in children_iter:
            if c.token == commands.CLOSING_BRACKET:
                break
            elif c.token not in "lcr|":
                raise InvalidAlignmentError
            alignment += c.token
        children = tuple(children_iter)
    elif (
        len(children)
        and children[0].children is not None
        and children[0].token == commands.BRACES
        and all(c.token in "lcr|" for c in children[0].children)
    ):
        alignment = "".join(c.token for c in children[0].children)
        children = children[1:]

    return Node(token=rf"\{environment}", children=children, alignment=alignment)


def _consume_brace_arg(tokens: Iterator[str]) -> list[str]:
    token = next(tokens)
    if token == "{":
        return _read_until_close_brace(tokens)
    return [token]


def _parse_optional_int(tokens: Iterator[str]) -> tuple[int, str]:
    peek = next(tokens)
    if peek != "[":
        return 0, peek
    nargs_str = ""
    for t in tokens:
        if t == "]":
            break
        nargs_str += t
    return int(nargs_str), next(tokens)


def _parse_newcommand(tokens: Iterator[str], macros: dict[str, tuple[list[str], int]]) -> None:
    name_tokens = _consume_brace_arg(tokens)
    name = "".join(name_tokens)
    nargs, peek = _parse_optional_int(tokens)
    if peek == "[":
        for t in tokens:
            if t == "]":
                break
        peek = next(tokens)
    if peek == "{":
        body = _read_until_close_brace(tokens)
    else:
        body = [peek]
    macros[name] = (body, nargs)


def _read_until_close_brace(tokens: Iterator[str]) -> list[str]:
    depth = 1
    content: list[str] = []
    for t in tokens:
        if t == "{":
            depth += 1
            content.append(t)
        elif t == "}":
            depth -= 1
            if depth == 0:
                return content
            content.append(t)
        else:
            content.append(t)
    return content


def _parse_newenvironment(tokens: Iterator[str], macros: dict[str, tuple[list[str], int]]) -> None:
    name_tokens = _consume_brace_arg(tokens)
    name = "".join(name_tokens)
    nargs, peek = _parse_optional_int(tokens)
    if peek == "{":
        begin_body = _read_until_close_brace(tokens)
    else:
        begin_body = [peek]
    end_body = _consume_brace_arg(tokens)
    macros[f"\\begin{{{name}}}"] = (begin_body, nargs)
    macros[f"\\end{{{name}}}"] = (end_body, 0)


def _parse_def(tokens: Iterator[str], macros: dict[str, tuple[list[str], int]]) -> None:
    name = next(tokens)
    nargs = 0
    for t in tokens:
        if t == "#":
            param = next(tokens, "")
            if param.isdigit():
                nargs = max(nargs, int(param))
        elif t == "{":
            break
    body = _read_until_close_brace(tokens)
    macros[name] = (body, nargs)


def _parse_declare_math_operator(tokens: Iterator[str], macros: dict[str, tuple[list[str], int]]) -> None:
    name_tokens = _consume_brace_arg(tokens)
    name = "".join(name_tokens)
    text_tokens = _consume_brace_arg(tokens)
    text = "".join(text_tokens)
    macros[name] = ([rf"\operatorname{{{text}}}"], 0)


def _substitute_params(body: list[str], args: list[list[str]]) -> list[str]:
    if not args:
        return list(body)
    expanded: list[str] = []
    body_iter = iter(body)
    for tok in body_iter:
        if tok == "#":
            param_num = next(body_iter, "")
            if param_num.isdigit() and 1 <= int(param_num) <= len(args):
                expanded.extend(args[int(param_num) - 1])
            else:
                expanded.append(tok)
                if param_num:
                    expanded.append(param_num)
        else:
            expanded.append(tok)
    return expanded


def _expand_macro(token: str, tokens: Iterator[str], macros: dict[str, tuple[list[str], int]]) -> list[str]:
    body, nargs = macros[token]
    if nargs == 0:
        return list(body)
    args: list[list[str]] = []
    for _ in range(nargs):
        args.append(_consume_brace_arg(tokens))
    return _substitute_params(body, args)



import re
from typing import Iterator

from latex2mathml import commands
from latex2mathml.symbols_parser import convert_symbol

UNITS = ("in", "mm", "cm", "pt", "em", "ex", "pc", "bp", "dd", "cc", "sp", "mu")

PATTERN = re.compile(
    rf"""
    (?P<comment>%[^\n]+) |
    (?P<letter>[a-zA-Z]) |
    (?P<subsup_operator>[_^])(?P<subsup_digit>\d) |
    (?P<dimension>-?\d+(?:\.\d+)?\s*(?:{"|".join(UNITS)})) |
    (?P<number>\d+(?:\.\d+)?) |
    (?P<dot_decimal>\.\d*) |
    (?P<escaped>\\[\\\[\]{{}}\s!,:>;|_%#$&]) |
    (?P<begin_end>\\(?:begin|end)\s*{{[a-zA-Z]+\*?}}) |
    (?P<operatorname>\\operatorname(?:withlimits|\*)?\s*{{[a-zA-Z\s*]+\*?\s*}}) |
    (?P<text_cmd>\\(?:cla(?:p|ss)|color(?!box)|emph|fbox|hbox|href|llap|mbox|rlap|style
        |tag\*?|text(?:bf|color|it|md|normal|rm|sf|tt|up)?|underbar))\s*{{(?P<text_content>[^}}]*)}} |
    (?P<frac_cmd>\\[cdt]?frac)\s*(?P<frac_arg1>[.\d])\s*(?P<frac_arg2>[.\d])? |
    (?P<math_font>\\math(?!ring|bin|close|inner|op|open|ord|punct|rel|strut)
        [a-z]+)(?P<math_open>{{)(?P<math_arg>[a-zA-Z])(?P<math_close>}}) |
    (?P<verb>\\verb(?P<verb_delim>.)(?P<verb_content>.*?)(?P=verb_delim)) |
    (?P<command>\\[a-zA-Z]+) |
    (?P<char>\S)
    """,
    re.VERBOSE,
)


def tokenize(latex_string: str, skip_comments: bool = True) -> Iterator[str]:
    """
    Converts Latex string into tokens.

    :param latex_string: Latex string.
    :param skip_comments: Flag to skip comments (default=True).
    """
    for match in PATTERN.finditer(latex_string):
        tokens = tuple(filter(lambda x: x is not None, match.groups()))
        if tokens[0].startswith(commands.VERB):
            yield commands.VERB
            yield tokens[2]  # verb_content
            continue
        if tokens[0].startswith(commands.MATH) and tokens[0] not in commands.MATH_NON_FONT_COMMANDS:
            full_math = "".join(tokens)
            symbol = convert_symbol(full_math)
            if symbol:
                yield f"&#x{symbol};"
                continue
        for captured in tokens:
            if skip_comments and captured.startswith("%"):
                break
            if captured.endswith(UNITS) and captured[0:1].isdigit():
                yield captured.replace(" ", "")
                continue
            if captured.startswith((commands.BEGIN, commands.END, commands.OPERATORNAME)):
                yield "".join(captured.split(" "))
                continue
            yield captured
